import {Injectable} from '@angular/core'
import {AlertService} from './alert.service'
import {AppState} from '../app-state'

@Injectable({
  providedIn: 'root'
})
export class ErrorService {
  private static readonly UNKNOWN_ERROR_MESSAGE = 'Une erreur inconnue est survenue'
  private defaultHandledErrorModels: ErrorModel[] = [{
    filter: error => error.status === 0 && error.statusText === 'Unknown Error' || error.status === 502,
    consumer: () => this.appState.isServerOnline = false
  }, {
    filter: error => error.status === 400,
    consumer: error => console.error(error),
    messageProducer: () => 'Certaines informations dans la requête étaient invalides'
  }, {
    filter: error => error.status === 401,
    messageProducer: () => 'Vous devez être connecté pour effectuer cette action'
  }, {
    filter: error => error.status === 403,
    messageProducer: () => 'Vous n\'avez pas la permission d\'effectuer cette action'
  }, {
    filter: error => error.status === 404,
    messageProducer: () => 'La resource demandée n\'a pas été trouvée'
  }, {
    filter: error => error.status === 500,
    messageProducer: () => ErrorService.UNKNOWN_ERROR_MESSAGE
  }]

  private activeHandler: ErrorHandler

  constructor(
    private alertService: AlertService,
    private appState: AppState
  ) {
  }

  handleError(error: any) {
    if (!this.activeHandler) {
      console.error('An error occurred but no handler was set')
    }

    let matchingModels = this.activeHandler.handledErrorModels.filter(m => m.filter(error)) // Find error models whose filter matches the current error

    if (!matchingModels || matchingModels.length == 0) {  // If none are found, search in defaults handlers
      matchingModels = this.defaultHandledErrorModels.filter(m => m.filter(error))
    }

    if (!matchingModels || matchingModels.length == 0) { // If still none are found, handle as an unknown error
      this.consumeUnknownError(error)
      return
    }

    matchingModels.forEach(m => {
      if (m.consumer || m.messageProducer) {
        if (m.consumer) {
          m.consumer(error)
        }
        if (m.messageProducer) {
          this.alertService.pushError(m.messageProducer(error))
        }
      } else {
        console.error('An error model has no consumer or message')
      }
    })
  }

  consumeUnknownError(error: any) {
    console.error(error)
    this.alertService.pushError(ErrorService.UNKNOWN_ERROR_MESSAGE)
  }

  set activeErrorHandler(handler: ErrorHandler) {
    this.activeHandler = handler
  }
}

/**
 * An error handler, defining models of errors that this type will handle
 */
export abstract class ErrorHandler {
  abstract handledErrorModels: ErrorModel[]

  protected constructor(
    protected errorService: ErrorService
  ) {
    this.errorService.activeErrorHandler = this
  }
}

/**
 * An error model define how errors matching its filter will be handled.
 *
 * The consumer will consume matching errors when they occurs.
 * The message producer returns a string that will be pushed to the alert system.
 *
 * To work correctly a model must define at least one handler (consumer or message producer).
 */
export class ErrorModel {
  constructor(
    public filter: (error: any) => Boolean,
    public consumer?: (error: any) => void,
    public messageProducer?: (error: any) => String
  ) {
  }
}
